@@ -1,23 +1,21 @@
+#![allow(unknown_lints)]
 use serde::{
-    de::{DeserializeSeed, Deserializer, SeqAccess, Visitor},
+    de::{DeserializeSeed, Deserializer, Error, SeqAccess, Visitor},
     ser::{Serialize, SerializeTupleStruct, Serializer},
 };
-use static_assertions::assert_impl_all;
-use std::convert::TryInto;
+use std::fmt::{Display, Write};
 
 use crate::{
-    signature_parser::SignatureParser, value::SignatureSeed, DynamicDeserialize, DynamicType,
-    OwnedValue, Signature, Type, Value,
+    DynamicDeserialize, DynamicType, OwnedValue, Signature, Value, value::SignatureSeed,
+    value_display_fmt,
 };
 
 /// Use this to efficiently build a [`Structure`].
 ///
 /// [`Structure`]: struct.Structure.html
-#[derive(Debug, Default, Clone, PartialEq)]
+#[derive(Debug, Default, PartialEq)]
 pub struct StructureBuilder<'a>(Vec<Value<'a>>);
 
-assert_impl_all!(StructureBuilder<'_>: Send, Sync, Unpin);
-
 impl<'a> StructureBuilder<'a> {
     /// Create a new `StructureBuilder`.
     ///
@@ -30,6 +28,7 @@ impl<'a> StructureBuilder<'a> {
     ///
     /// This method returns `Self` so that you can use the builder pattern to create a complex
     /// structure.
+    #[must_use]
     pub fn add_field<T>(self, field: T) -> Self
     where
         T: DynamicType + Into<Value<'a>>,
@@ -40,6 +39,7 @@ impl<'a> StructureBuilder<'a> {
     /// Append `field` to `self`.
     ///
     /// Identical to `add_field`, except the field must be in the form of a `Value`.
+    #[must_use]
     pub fn append_field<'e: 'a>(mut self, field: Value<'e>) -> Self {
         self.0.push(field);
 
@@ -66,69 +66,86 @@ impl<'a> StructureBuilder<'a> {
     /// Build the `Structure`.
     ///
     /// [`Structure`]: struct.Structure.html
-    pub fn build(self) -> Structure<'a> {
-        let signature = create_signature_from_fields(&self.0);
+    pub fn build(self) -> crate::Result<Structure<'a>> {
+        if self.0.is_empty() {
+            return Err(crate::Error::EmptyStructure);
+        }
 
-        Structure {
+        let fields_signatures: Box<[Signature]> =
+            self.0.iter().map(Value::value_signature).cloned().collect();
+        let signature = Signature::structure(fields_signatures);
+
+        Ok(Structure {
             fields: self.0,
             signature,
-        }
+        })
     }
 
     /// Same as `build` except Signature is provided.
-    pub(crate) fn build_with_signature<'s: 'a>(self, signature: Signature<'s>) -> Structure<'a> {
+    pub(crate) fn build_with_signature<'s: 'a>(self, signature: &Signature) -> Structure<'a> {
         Structure {
             fields: self.0,
-            signature,
+            signature: signature.clone(),
         }
     }
 }
 
 /// Use this to deserialize a [`Structure`].
 ///
-/// [`Structure`]: struct.Structure.html
-#[derive(Debug, Clone, PartialEq)]
-pub struct StructureSeed<'a>(Signature<'a>);
+/// The lifetime `'a` is now redundant and kept only for backward compatibility. All instances now
+/// has a `'static` lifetime. This will be removed in the next major release.
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct StructureSeed<'a> {
+    signature: Signature,
+    phantom: std::marker::PhantomData<&'a ()>,
+}
 
-assert_impl_all!(StructureSeed<'_>: Send, Sync, Unpin);
-
-impl<'a> StructureSeed<'a> {
+impl StructureSeed<'static> {
     /// Create a new `StructureSeed`
     ///
     /// The given signature must be a valid structure signature.
-    pub fn new_unchecked(signature: Signature<'a>) -> Self {
-        StructureSeed(signature)
-    }
-}
-
-impl<'a> std::convert::TryFrom<Signature<'a>> for StructureSeed<'a> {
-    type Error = zvariant::Error;
-
-    fn try_from(signature: Signature<'a>) -> Result<Self, zvariant::Error> {
-        if signature.starts_with(zvariant::STRUCT_SIG_START_CHAR) {
-            Ok(StructureSeed(signature))
-        } else {
-            Err(zvariant::Error::IncorrectType)
+    #[must_use]
+    pub fn new_unchecked(signature: &Signature) -> Self {
+        StructureSeed {
+            signature: signature.clone(),
+            phantom: std::marker::PhantomData,
         }
     }
 }
 
-impl<'de> DeserializeSeed<'de> for StructureSeed<'de> {
+impl TryFrom<Signature> for StructureSeed<'static> {
+    type Error = zvariant::Error;
+
+    fn try_from(signature: Signature) -> Result<Self, zvariant::Error> {
+        if !matches!(signature, Signature::Structure(_)) {
+            return Err(zvariant::Error::IncorrectType);
+        }
+
+        Ok(StructureSeed {
+            signature,
+            phantom: std::marker::PhantomData,
+        })
+    }
+}
+
+impl<'de> DeserializeSeed<'de> for StructureSeed<'_> {
     type Value = Structure<'de>;
     fn deserialize<D>(self, deserializer: D) -> Result<Self::Value, D::Error>
     where
         D: Deserializer<'de>,
     {
-        deserializer.deserialize_seq(StructureVisitor { signature: self.0 })
+        deserializer.deserialize_seq(StructureVisitor {
+            signature: self.signature,
+        })
     }
 }
 
-#[derive(Debug, Clone, PartialEq)]
-struct StructureVisitor<'a> {
-    signature: Signature<'a>,
+#[derive(Debug, Clone, PartialEq, Eq)]
+struct StructureVisitor {
+    signature: Signature,
 }
 
-impl<'de> Visitor<'de> for StructureVisitor<'de> {
+impl<'de> Visitor<'de> for StructureVisitor {
     type Value = Structure<'de>;
 
     fn expecting(&self, formatter: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
@@ -140,7 +157,7 @@ impl<'de> Visitor<'de> for StructureVisitor<'de> {
         V: SeqAccess<'de>,
     {
         SignatureSeed {
-            signature: self.signature,
+            signature: &self.signature,
         }
         .visit_struct(visitor)
     }
@@ -151,14 +168,12 @@ impl<'de> Visitor<'de> for StructureVisitor<'de> {
 /// API is provided to convert from, and to tuples.
 ///
 /// [`Value`]: enum.Value.html
-#[derive(Debug, Clone, PartialEq)]
+#[derive(Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
 pub struct Structure<'a> {
     fields: Vec<Value<'a>>,
-    signature: Signature<'a>,
+    signature: Signature,
 }
 
-assert_impl_all!(Structure<'_>: Send, Sync, Unpin);
-
 impl<'a> Structure<'a> {
     /// Get a reference to all the fields of `self`.
     pub fn fields(&self) -> &[Value<'a>] {
@@ -170,126 +185,117 @@ impl<'a> Structure<'a> {
         self.fields
     }
 
-    /// Create a new `Structure`.
-    ///
-    /// Same as `Structure::default()`.
-    #[deprecated(
-        since = "2.3.0",
-        note = "Please use `StructureBuilder` to create a `Structure` instead."
-    )]
-    pub fn new() -> Self {
-        Self::default()
-    }
-
-    /// Append `field` to `self`.
-    ///
-    /// This method returns `Self` so that you can use the builder pattern to create a complex
-    /// structure.
-    #[deprecated(
-        since = "2.3.0",
-        note = "Please use `StructureBuilder` to create a `Structure` instead."
-    )]
-    pub fn add_field<T>(self, field: T) -> Self
-    where
-        T: Type + Into<Value<'a>>,
-    {
-        #[allow(deprecated)]
-        self.append_field(Value::new(field))
-    }
-
-    /// Append `field` to `self`.
-    ///
-    /// Identical to `add_field`, except the field must be in the form of a `Value`.
-    #[deprecated(
-        since = "2.3.0",
-        note = "Please use `StructureBuilder` to create a `Structure` instead."
-    )]
-    pub fn append_field<'e: 'a>(mut self, field: Value<'e>) -> Self {
-        self.fields.push(field);
-        self.signature = create_signature_from_fields(&self.fields);
-
-        self
-    }
-
     /// Get the signature of this `Structure`.
-    ///
-    /// NB: This method potentially allocates and copies. Use [`full_signature`] if you'd like to
-    /// avoid that.
-    ///
-    /// [`full_signature`]: #method.full_signature
-    pub fn signature(&self) -> Signature<'static> {
-        self.signature.to_owned()
-    }
-
-    /// Get the signature of this `Structure`.
-    pub fn full_signature(&self) -> &Signature<'_> {
+    pub fn signature(&self) -> &Signature {
         &self.signature
     }
 
-    pub(crate) fn to_owned(&self) -> Structure<'static> {
-        Structure {
-            fields: self.fields.iter().map(|v| v.to_owned()).collect(),
+    pub(crate) fn try_to_owned(&self) -> crate::Result<Structure<'static>> {
+        Ok(Structure {
+            fields: self
+                .fields
+                .iter()
+                .map(|v| v.try_to_owned().map(Into::into))
+                .collect::<crate::Result<_>>()?,
             signature: self.signature.to_owned(),
-        }
+        })
+    }
+
+    pub(crate) fn try_into_owned(self) -> crate::Result<Structure<'static>> {
+        Ok(Structure {
+            fields: self
+                .fields
+                .into_iter()
+                .map(|v| v.try_into_owned().map(Into::into))
+                .collect::<crate::Result<_>>()?,
+            signature: self.signature,
+        })
+    }
+
+    /// Attempt to clone `self`.
+    pub fn try_clone(&self) -> Result<Self, crate::Error> {
+        let fields = self
+            .fields
+            .iter()
+            .map(|v| v.try_clone())
+            .collect::<crate::Result<Vec<_>>>()?;
+
+        Ok(Self {
+            fields,
+            signature: self.signature.clone(),
+        })
     }
 }
 
-impl<'a> Default for Structure<'a> {
-    fn default() -> Self {
-        let signature = Signature::from_static_str_unchecked("()");
-
-        Self {
-            fields: vec![],
-            signature,
-        }
+impl Display for Structure<'_> {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        structure_display_fmt(self, f, true)
     }
 }
 
-impl<'a> DynamicType for Structure<'a> {
-    fn dynamic_signature(&self) -> Signature<'_> {
+pub(crate) fn structure_display_fmt(
+    structure: &Structure<'_>,
+    f: &mut std::fmt::Formatter<'_>,
+    type_annotate: bool,
+) -> std::fmt::Result {
+    f.write_char('(')?;
+
+    let fields = structure.fields();
+
+    match fields.len() {
+        0 => {}
+        1 => {
+            value_display_fmt(&fields[0], f, type_annotate)?;
+            f.write_char(',')?;
+        }
+        _ => {
+            for (i, field) in fields.iter().enumerate() {
+                value_display_fmt(field, f, type_annotate)?;
+
+                if i + 1 < fields.len() {
+                    f.write_str(", ")?;
+                }
+            }
+        }
+    }
+
+    f.write_char(')')
+}
+
+impl DynamicType for Structure<'_> {
+    fn signature(&self) -> Signature {
         self.signature.clone()
     }
 }
 
-impl<'a> DynamicType for StructureSeed<'a> {
-    fn dynamic_signature(&self) -> Signature<'_> {
-        self.0.clone()
+impl DynamicType for StructureSeed<'_> {
+    fn signature(&self) -> Signature {
+        self.signature.clone()
     }
 }
 
 impl<'a> DynamicDeserialize<'a> for Structure<'a> {
-    type Deserializer = StructureSeed<'a>;
+    type Deserializer = StructureSeed<'static>;
 
-    fn deserializer_for_signature<S>(signature: S) -> zvariant::Result<Self::Deserializer>
-    where
-        S: TryInto<Signature<'a>>,
-        S::Error: Into<zvariant::Error>,
-    {
-        let mut signature = signature.try_into().map_err(Into::into)?;
-        if !signature.starts_with(zvariant::STRUCT_SIG_START_CHAR) {
-            // This is certainly not a valid struct signature
-            signature = format!("({})", signature).try_into()?;
-            return signature.try_into();
-        }
+    fn deserializer_for_signature(signature: &Signature) -> zvariant::Result<Self::Deserializer> {
+        let signature = match signature {
+            Signature::Structure(_) => signature.clone(),
+            s => Signature::structure([s.clone()]),
+        };
 
-        // The signature might be something like "(i)u(i)" - we need to parse it to check.
-        let mut parser = SignatureParser::new(signature.clone());
-        parser.parse_next_signature()?;
-        if !parser.done() {
-            // more than one element - we must wrap it
-            signature = format!("({})", signature).try_into()?;
-        }
-        signature.try_into()
+        Ok(StructureSeed {
+            signature,
+            phantom: std::marker::PhantomData,
+        })
     }
 }
 
-impl<'a> Serialize for Structure<'a> {
+impl Serialize for Structure<'_> {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
         S: Serializer,
     {
-        let mut structure =
-            serializer.serialize_tuple_struct("zvariant::Structure", self.fields.len())?;
+        let mut structure = serializer.serialize_tuple_struct("Structure", self.fields.len())?;
         for field in &self.fields {
             field.serialize_value_as_tuple_struct_field(&mut structure)?;
         }
@@ -310,13 +316,13 @@ macro_rules! tuple_impls {
                     $(
                         .add_field(value. $n)
                     )+
-                    .build()
+                    .build().unwrap()
                 }
             }
 
-            impl<'a, E, $($name),+> std::convert::TryFrom<Structure<'a>> for ($($name),+,)
+            impl<'a, E, $($name),+> TryFrom<Structure<'a>> for ($($name),+,)
             where
-                $($name: std::convert::TryFrom<Value<'a>, Error = E>,)+
+                $($name: TryFrom<Value<'a>, Error = E>,)+
                 crate::Error: From<E>,
 
             {
@@ -331,9 +337,22 @@ macro_rules! tuple_impls {
                 }
             }
 
-            impl<'a, E, $($name),+> std::convert::TryFrom<Value<'a>> for ($($name),+,)
+            impl<'a, E, $($name),+> TryFrom<&Value<'a>> for ($($name),+,)
             where
-                $($name: std::convert::TryFrom<Value<'a>, Error = E>,)+
+                $($name: TryFrom<Value<'a>, Error = E>,)+
+                crate::Error: From<E>,
+
+            {
+                type Error = crate::Error;
+
+                fn try_from(v: &Value<'a>) -> core::result::Result<Self, Self::Error> {
+                    Self::try_from(Structure::try_from(v)?)
+                }
+            }
+
+            impl<'a, E, $($name),+> TryFrom<Value<'a>> for ($($name),+,)
+            where
+                $($name: TryFrom<Value<'a>, Error = E>,)+
                 crate::Error: From<E>,
 
             {
@@ -344,9 +363,9 @@ macro_rules! tuple_impls {
                 }
             }
 
-            impl<E, $($name),+> std::convert::TryFrom<OwnedValue> for ($($name),+,)
+            impl<E, $($name),+> TryFrom<OwnedValue> for ($($name),+,)
             where
-                $($name: std::convert::TryFrom<Value<'static>, Error = E>,)+
+                $($name: TryFrom<Value<'static>, Error = E>,)+
                 crate::Error: From<E>,
 
             {
@@ -379,13 +398,52 @@ tuple_impls! {
     16 => (0 T0 1 T1 2 T2 3 T3 4 T4 5 T5 6 T6 7 T7 8 T8 9 T9 10 T10 11 T11 12 T12 13 T13 14 T14 15 T15)
 }
 
-fn create_signature_from_fields(fields: &[Value<'_>]) -> Signature<'static> {
-    let mut signature = String::with_capacity(255);
-    signature.push('(');
-    for field in fields {
-        signature.push_str(&field.value_signature());
-    }
-    signature.push(')');
+/// Owned [`Structure`]
+#[derive(Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
+pub struct OwnedStructure(pub Structure<'static>);
 
-    Signature::from_string_unchecked(signature)
+/// Use this to deserialize an [`OwnedStructure`].
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct OwnedStructureSeed(Signature);
+
+impl DynamicType for OwnedStructure {
+    fn signature(&self) -> Signature {
+        self.0.signature().clone()
+    }
+}
+
+impl DynamicType for OwnedStructureSeed {
+    fn signature(&self) -> Signature {
+        self.0.clone()
+    }
+}
+
+impl DynamicDeserialize<'_> for OwnedStructure {
+    type Deserializer = OwnedStructureSeed;
+
+    fn deserializer_for_signature(signature: &Signature) -> zvariant::Result<Self::Deserializer> {
+        Structure::deserializer_for_signature(signature)
+            .map(|StructureSeed { signature, .. }| OwnedStructureSeed(signature))
+    }
+}
+
+impl<'de> DeserializeSeed<'de> for OwnedStructureSeed {
+    type Value = OwnedStructure;
+    fn deserialize<D: Deserializer<'de>>(self, deserializer: D) -> Result<Self::Value, D::Error> {
+        deserializer
+            .deserialize_seq(StructureVisitor { signature: self.0 })
+            .and_then(|s| match s.try_to_owned() {
+                Ok(s) => Ok(OwnedStructure(s)),
+                Err(e) => Err(D::Error::custom(e)),
+            })
+    }
+}
+
+impl Serialize for OwnedStructure {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        self.0.serialize(serializer)
+    }
 }
